// filter -m minSize -M maxSize -e extensions file-list
// extension = comma separated extensions, including .

package main

import (
	"flag"
	"fmt"
	"os"
	"path/filepath"
	"runtime"
	"strings"
)

func main() {
	// Set maximum used CPUs to the available CPUs
	runtime.GOMAXPROCS(runtime.NumCPU())

	// get the list of desired suffixes and the files
	minSize, maxSize, suffixes, files := handleCommandLine()

	// chaining
	sink(
		filterSize(minSize, maxSize,
			filterSuffixes(suffixes,
				source(files))))
}

func handleCommandLine() (minSize, maxSize int64, suffixes map[string]bool, files []string) {
	// flags -m and -M
	flag.Int64Var(&minSize, "m", -1,
		"minimum file size (-1 means no minimum)")
	flag.Int64Var(&maxSize, "M", -1,
		"maximum file size (-1 means no maximum)")

	// flag -e
	var suffixesOpt *string = flag.String("e", "",
		"comma-separated list of file suffixes")

	// parse the command line
	flag.Parse()

	// turn suffixes opt into a slice
	suffixes = make(map[string]bool)
	if *suffixesOpt != "" {
		for _, suff := range strings.Split(*suffixesOpt, ",") {
			suffixes[suff] = true
		}
	}

	// get the file names
	files = flag.Args()

	return minSize, maxSize, suffixes, files
}

// feed the chain with file names
func source(files []string) <-chan string {
	out := make(chan string, 1000)
	go func() {
		for _, filename := range files {
			out <- filename
		}
		close(out)
	}()
	return out
}


func filterSuffixes(suffixes map[string]bool, in <-chan string) <-chan string {
	// output channel (same size as input to maximize throughput)
	out := make(chan string, cap(in))
	go func() {
		for filename := range in {
			// if no suffixes specified (continue?)
			if len(suffixes) == 0 {
				out <- filename				
				continue
			}

			ext := strings.ToLower(filepath.Ext(filename))
			if _, present := suffixes[ext]; present {
				out <- filename
			}
		}
		close(out)
	}()
	return out
}


func filterSize(minimum, maximum int64, in <-chan string) <-chan string {
	// output channel (same size as input to maximize throughput)
	out := make(chan string, cap(in))
	go func() {
		for filename := range in {
			// can be optimised avoiding stat call, when not needed

			finfo, err := os.Stat(filename)
			// ignore files we can't process
			if err == nil {
				size := finfo.Size()
				if (minimum == -1 || minimum > -1 && minimum <= size) &&
					(maximum == -1 || maximum > -1 && maximum >= size) {
					out <- filename
				}
			}
		}
		close(out)
	}()
	return out
}

func sink(in <-chan string) {
	for filename := range in {
		fmt.Println(filename)
	}
}
