/* Computing and imaging Mandelbrot set */


package mandelpkg


import "math"
import "image"
import "image/jpeg"
import "image/color"
import "os"
import "math/cmplx"
import "fmt"
import "bufio"


/* Checks whether z_n, defined by z_0=0 and z_n+1 = z_n^2 + z tends to
   infinity. Concretely, one checks whether the module becomes greater
   than 2, and and returns. 

   The number of iterations and, for the same number of iterations,§
   the module of the last element of the series is as measure of the
   divergence speed.  
*/

func check(z complex128, maxIters int) float64 {	
	var iters int = 0
	var zn complex128 = complex(0,0)
	
	var len float64 = 0
	for ((len <= 2) && (iters < maxIters)) {
		zn = zn*zn + z
		len = cmplx.Abs(zn)		
		iters++
	}

	// the speed is a value in [0,1] determined by the number of
	// iterations, corrected with abs(z) to have a smoother
	// variation
	var speed float64
	if ( iters < maxIters ) {
		speed = float64(iters + 1) - math.Log10(math.Log2(cmplx.Abs(zn)))/float64(maxIters)
	} else {
		speed = 0
	}
	
	return speed
}


/* functions for image managing */

/* colored point */

type ColoredPoint struct {
	i, j   int
	col    color.Color
}

// constructor of colored point
func  NewColoredPoint (i, j int, col color.Color) ColoredPoint {
	return ColoredPoint{i,j,col}
}

// method for getting the trhree components of a colored point
func (self *ColoredPoint)  GetComponents () (int, int, color.Color) {
	return self.i, self.j, self.col
}


// convert pixel positions i,j to float coordinates
func convert(i,j, imgW, imgH int, x0, y0, offx, offy float64) (float64, float64) {
	x1 := x0 + (offx * float64(i))/float64(imgW)
	y1 := y0 + (offy * float64(j))/float64(imgH)
	return x1, y1
	       
}

// color a point according to the value of the speed
func ElaboratePoint(point *ColoredPoint, imgW, imgH, maxIters int, x0, y0, offx, offy float64, colorIn, colorOut color.RGBA) {
	x,y := convert(point.i,point.j, imgW, imgH, x0, y0, offx, offy)
	z:= complex(x, y)
	speed := check(z,maxIters)
	point.col = chooseColor(speed, colorIn, colorOut)
}


/* choose a color, interpolating between two given colors using the speed */

func chooseColor(speed float64, start, end color.RGBA) color.RGBA {
	
	red   := uint8(float64(start.R + (end.R - start.R))*speed)
	green := uint8(float64(start.G + (end.G - start.G))*speed)
	blue  := uint8(float64(start.B + (end.B - start.B))*speed)
	
	return color.RGBA{red, green, blue, 0xFF}
}


func WriteImage(img *image.RGBA, filename string) {
	file, err := os.Create(filename + ".jpg")
	if err != nil {
		fmt.Printf("Could not create file %s", file.Name())
	}
	writer := bufio.NewWriter(file)
	quality := jpeg.Options{80}
	jpeg.Encode(writer, img, &quality)
	writer.Flush()
	file.Close()
}



