/* perform the same computation on a set of elements
   version with flags for specifying
   - number of elements to elaborate
   - type of taks (IO or CPU bound)
   - concurrent or sequential mode
*/

package main

import (
	"fmt"
	"io/ioutil"
	"net/http"
	"flag"
	)

/* REQUEST *****************************************************/

// size for buffering requests
const BUF_SIZE = 100000

type Operator func (int) int

type Request struct {
	Op        Operator  // operator to be applied
	InputChan chan int  // chan to input elements to be processed 
	ReplyChan chan int  // chan where processed elements are returned
}

/* SERVER *****************************************************/

// parametrised by an input port for receiving
func server(service <-chan *Request, concurrent bool) {
	for {
		req := <-service    // requests arrive here 
		go serve(req, concurrent)       // don't wait for op
	}
}

// function serving the request: it inputs all elements from
// InputChan, create a process for each input, collect the results,
// send them to the ReplyChannel and close it
func serve(req *Request, concurrent bool) {
	// create a channel for waiting slaves to complete
	done := make(chan bool, BUF_SIZE)
	numSlaves := 0
	for val := range req.InputChan {
		// concurrent or non-concurrent, this is the problem
		if concurrent {
			go slave(req.Op, val, req.ReplyChan, done)
		} else {
			slave(req.Op, val, req.ReplyChan, done)
		}
		numSlaves++
	}
	for (numSlaves > 0) {
		<-done
		numSlaves--
	}
	close(req.ReplyChan)
}


// slave function for processing a single element
func slave(op Operator, val int, reply chan<- int, done chan<- bool) {
	reply <- op(val)
	done <- true
}
	

// function for starting the server: returns a port for sending requests
func startServer(concurrent bool) chan<- *Request {
	service := make(chan *Request) 
	go server(service, concurrent) 
	return service
}

// do some lengthy, not CPU intensive work: IO over network
func useIO (n int) int {	
	url := "https://it.wikipedia.org/wiki/Italia"

	// send a get request (ignore a possible error)
	resp, err := http.Get(url)
	if (err != nil) {
		fmt.Println("Some error in the get request!")
		return 0
	} else {
		// close the response body
		defer resp.Body.Close()
		
		// read html as a slice of bytes (ignore a possible error)
		html, errnew := ioutil.ReadAll(resp.Body)
		if (errnew != nil) {
			fmt.Println("Some error in getting the body!")
		}
		return len(html)
	}
}


// do some CPU intensive task
func useCPU (n int) int {
	sum := 0 
	for i:=0; i< n; i++ {
		for j:=0; j< n; j++ {
			sum = sum + i*j - i*j
		}
	}
	return sum
}

func main() {
	// number of elements to process
	var numElems int

	// choose CPU/IO bound tasks
	var cpu bool

	// concurrent or not
	var concurrent bool
	
	flag.IntVar(&numElems, "num", 100, "length of stream of elements to process ")
	flag.BoolVar(&cpu, "cpu", false, "execute a CPU bound task (default IO bound)")
	flag.BoolVar(&concurrent, "conc", false, "execute slaves concurrently (default sequentially)")

	flag.Parse()

	fmt.Println(numElems, " elements")
	if cpu {
		fmt.Println("CPU bound task")
	} else {
		fmt.Println("IO bound task")
	}
	if concurrent {
		fmt.Println("Concurrent")
	} else {
		fmt.Println("Sequential")
	}

	// create a server instance
	serverChan := startServer(concurrent)
	
	// client side
	var op Operator
	if cpu {
		op = useCPU
	} else {
		op = useIO
	}
	
	req := &Request{
		op,
		make(chan int, BUF_SIZE),
		make(chan int, BUF_SIZE)} 


	// send data on inputChannel
	serverChan <- req

	for val := 0; val < numElems; val++ {
		req.InputChan <- val
	}
	close(req.InputChan)
	
	// retrieve the results ranging on the reply channel
	sum := 0
	for v := range req.ReplyChan {
		sum += v
	}

	fmt.Printf("Request out: %d\n", sum)
}
