package main

import ( 
	"fmt" 
 	"time"
	"math/rand"
	)


// empty structs used for signalling on channels
type Signal struct{}
var sigVal = struct{}{}

// Number of philosophers
const NPhil = 5

// Start value of the counters
const MaxToken = 5


// Fork type
type Fork struct {
	name string      //  fork name

	// (public) channels
	Take chan Signal     //  request of taking
	Leave chan Signal     //  release
}


// Fork behaviour
func (self *Fork) Run () {
	
	for {
		<- self.Take
		fmt.Printf("Fork %s taken\n", self.name) 
		<- self.Leave
		fmt.Printf("Fork %s released\n", self.name)      
	}	
}


// Counter type
type Counter struct {
	// top value
	top int

	// channels
	Dec chan Signal       //  use a token
	Res chan Signal       //  restore
}

// Counter behaviour
func (self *Counter) Run () {
	var curVal int
	for {
		curVal = self.top
		for curVal > 0 { // if there are tokens available, provide one
			<- self.Dec
			curVal = curVal- 1
		}
		// when value goes down to zero, wait to be reset
		<- self.Res 
	}
}


/* resetter (takes in input the array) */
func Resetter (counters []Counter) {
	for {
		for _,counter := range counters {
			counter.Res <- sigVal
		}
	}
}


// Philosopher behaviour

func Phil (id int, left *Fork, right *Fork, counter *Counter) {

	/* number of tokens disregarded consecutively */
	disregarded := 0

	for {	
		/* decrement the counter, if possible */
		counter.Dec <- sigVal
		
		/* the probability of eating increases with hunger */
		if (disregarded < MaxToken) && (rand.Intn(disregarded+1) == 0) { 
			disregarded++
			fmt.Printf("Philosopher %d, token disregarded ... %d times\n", id, disregarded)			
		} else {
			fmt.Printf("Philosopher %d is thinking ...\n", id)
			/* sleep for at most 2 seconds */
			time.Sleep(time.Duration(rand.Int63n(2*1e9)))
			/* take left fork */
			left.Take <- sigVal
			/* sleep for at most 2 seconds */
			time.Sleep(time.Duration(rand.Int63n(2*1e9)))
			/* take right fork */
			right.Take <- sigVal
			fmt.Printf("Philosopher %d is eating ...\n", id)
			left.Leave <- sigVal
			right.Leave <- sigVal

			// starts again, reset disregarded tokens
			disregarded = 0
		}
		
	}
}



// main
func main () {
	var forks [NPhil]Fork
	var counters [NPhil]Counter
	var i int
	
	for i=0; i<NPhil; i++ {

		// create the ith fork
		forks[i] = Fork { 
			fmt.Sprintf("F%d",i), 
			make(chan Signal), 
			make(chan Signal)}

		go forks[i].Run()

		// create the ith counter
		counters[i] = Counter { 
			MaxToken, 
			make(chan Signal), 
			make(chan Signal)}

		go counters[i].Run()		
	}
	

	/* start the resetter */
	go Resetter(counters[:])


	/* start the philosophers */
	for i=0; i<NPhil-1; i++ {
		go Phil(i, &forks[i], &forks[(i+1)%NPhil], &counters[i])
	}
		
	// left-handed last philospher: no deadlock!
	Phil(NPhil-1, &forks[0], &forks[NPhil-1], &counters[NPhil-1])

	// right-handed last philospher: deadlock!
	// Phil(NPhil-1, &forks[NPhil-1], &forks[0], &counters[NPhil-1])
			
}

