// channels of channels: a server processing requests and returning
// the result on a dedicated channel

package main

import (
	"fmt"
	"math/rand"
	)

/* REQUEST *****************************************************/
// the client send a request, with operands a and b and a channel
// where to answer.

type Request struct { 
	a,b int
	replyChan chan int
}

// function which waits for the answer to the request on the channel
// and returns it
func (r *Request) GetResult() int { 
	return <-r.replyChan
}

// function for getting the parameters of the request
func (r *Request) GetPars() (int, int) { 
	return r.a, r.b
}

// function for sending the result on the reply channel
func (r *Request) SendRes(val int) { 
	r.replyChan <- val
}


/* SERVER *****************************************************/
// the type of a possible operator to apply (functions are first-class
// values)

type binOp func(a, b int) int


// server: parametrised by an operation to be executed and an input
// port for receiving
func server(op binOp, service <-chan *Request) {
	for {
		req := <-service    // requests arrive here 
		go serve(op, req)   // don't wait for op
	}
}

// function serving the request: it applies the operator to the
// operands and returns the result in the desired channel
func serve(op binOp, req *Request) { 
	req.SendRes(op(req.GetPars()))
}


// function for starting the server: it provides the operator to apply
// and returns a port for sending requests
func StartServer(op binOp) chan<- *Request {
	service := make(chan *Request) 
	go server(op, service) 
	return service
}


func main() {

	// create a server instance, performing additions (unnamed functions)
	sum := func(a, b int) int { return a + b }
	serverChanSum := StartServer(sum)
	op := '+'

	// same for multiplication
	prod := func(a, b int) int { return a * b }
	serverChanProd := StartServer(prod)
	op1 := '*'


	// client side
	reqSum  := &Request{rand.Intn(100), rand.Intn(100), make(chan int)} 
	reqProd := &Request{rand.Intn(100), rand.Intn(100), make(chan int)}

	// send the requests
	serverChanSum  <- reqSum 
	serverChanProd <- reqProd
	
	// Can retrieve results in any order
	fmt.Printf("Request 2: %d %c  %d = %d\n",
		reqProd.a, 
		op1,
		reqProd.b,
		reqProd.GetResult()) 


	fmt.Printf("Request 1: %d %c %d = %d\n",
		reqSum.a, 
		op,
		reqSum.b,
		reqSum.GetResult()) 

}
