// server computing sums: works at http://localhost:4000/calc

package main

import (
	"fmt"
	"log"
	"net/http"   // values which implements http.Handler can be
		     // used to serve http requests
	"strconv"
	"strings"
	"errors"
)


const (
    pageTop    = 
    `<!DOCTYPE HTML><html><head>
     <style>.error{color:#FF0000;}</style></head>
     <title>Statistics</title>
     <body>
        <h3>Multiplier</h3>
        <p>Computes the product of some numbers (wow!)</p>`

    form       = 
    `<form action="/calc" method="POST">
       <label for="numbers">Numbers (space-separated):</label><br />
       <input type="text" name="numbers" size="30"><br />
       <input type="submit" value="Calculate">
     </form>`

    pageBottom = `</body></html>`

    anError    = `<p class="error">%s</p>`
)


func main() {
	/* define the handler for path "/calc" */
	http.HandleFunc("/calc", homePage)
	
	/* start the server (port 4000 of localhost) */
	if err := http.ListenAndServe(":4000", nil); err != nil {
		log.Fatal("failed to start server", err)
	}
}


/* the handler function must have the two parameters below
 -  writer: file id for writing
 -  request: map containing the passed parameters (type map[string][]string)
 */

func homePage(writer http.ResponseWriter, request *http.Request) {
	// Parse raw form (initially empty), possibly catching errors
	err := request.ParseForm() 

	// start the page (top and form)
	fmt.Fprint(writer, pageTop, form)
	
	if err != nil {
		/* some error occurred: out it (using the const
		   anError as pattern) */
		fmt.Fprintf(writer, anError, err)
	} else {
		if numbers, err := processRequest(request); err != nil {
			fmt.Fprintf(writer, anError, err)
		} else if len(numbers) > 0 {    // if there are some numbers to multiply
			fmt.Fprint(writer, formatOut(numbers))
		}
	}
	fmt.Fprint(writer, pageBottom)
}


/*
 process the request, getting the numbers in input (returns slice of
 input numbers) and an error
*/

func processRequest(request *http.Request) ([]int64, error){ 
	var numbers []int64

	/* get the "numbers" field, which is a slice with a single element */
	if slice, found := request.Form["numbers"]; found && len(slice) > 0 {
		
		/* Fields: split the input string into a slice of
		 strings (assuming  space as a separator) */
		for _, field := range strings.Fields(slice[0]) {
			/* ParseInt(string, base, bitsize) */
			if x, err := strconv.ParseInt(field, 10, 64); err != nil {
				/* if number is invalid, return with an error */
				return numbers, errors.New("'" + field + "' is invalid")
			} else {
				/* append the number to the slice */
				numbers = append(numbers, x)
			}
		}
	}
	// return the numbers (possibly no data if first time form is
	// shown or nothing specified) and indication of no error	
	return numbers, nil
}


func prod(numbers []int64) (product int64) {
	product = 1
	for _, x := range numbers {
		product *= x
	}
	return product
}

func formatOut(numbers []int64) string {
    return fmt.Sprintf(`
<table border="1">
	<tr><th colspan="2">Results</th></tr>
	<tr><td>Numbers</td><td>%v</td></tr>
	<tr><td>Product</td><td>%d</td></tr>
</table>`, numbers, prod(numbers))
}
